/* Kanoniczny program MPI. */
#include <stdio.h>
#include <string.h>
#include <mpi.h>

static int const NUM_INTEGRATION_INTERVALS = 1024 * 1024;

int main(int argc, char * argv[])
{
  int myProcessNo;
  int numProcesses;
  int coordinatorProcessNo = 0;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &myProcessNo);
  MPI_Comm_size(MPI_COMM_WORLD, &numProcesses);
  
  if (myProcessNo < NUM_INTEGRATION_INTERVALS)
  {
    /* Jedynie NUM_INTEGRATION_INTERVALS procesow oblicza PI. */
    double step;
    double sum;
    double myPartialIntegral;
    double totalIntegral = 0;
    int i;

    /* Oblicz krok calkowania i zainicjuj sume. */
    step = 1 / (double)NUM_INTEGRATION_INTERVALS;
    sum = 0.0;

    /* Policz calke na zadanym przedziale z krokiem h. */
    for (i = myProcessNo + 1; i <= NUM_INTEGRATION_INTERVALS; i += numProcesses)
    {
      double x = step * ((double)i - 0.5);  /* Metoda trapezow. */
      sum += 4.0 / (1.0 + x * x);
    }
    myPartialIntegral = sum * step;

    /* Oblicz sume sum czesciowych po wszystkich procesach. */
    MPI_Reduce(
        &myPartialIntegral,    /* co agregujemy */
        &totalIntegral,        /* gdzie przechowujemy wynik (pamiec) */
        1,                     /* ile elementow agregujemy */
        MPI_DOUBLE,            /* jakiego typu elementy agregujemy */
        MPI_SUM,               /* jakiej funkcji agregujacej uzywamy */
        coordinatorProcessNo,  /* gdzie przechowujemy wynik (proces) */
        MPI_COMM_WORLD
    );
    
    if (myProcessNo == coordinatorProcessNo)
    {
      printf("PI = %f.10\n", totalIntegral);
    }
  }

  MPI_Finalize();

  return 0;
}
